#include <stdio.h>
#include <sys/time.h>
#include <stdlib.h>
#include <setjmp.h>
#include <unistd.h>
#include <signal.h>
#include <readline/readline.h>
#include <readline/history.h>
#include "lac.h"

lenv_t *null_env;

void
lac_error_print (FILE * f)
{
  fprintf (f, "(*LAC-ERROR* \"%s", lac_errmsg ());
  if (lac_errlreg () != NIL)
    {
      fprintf (f, ": ");
      sexpr_fprint (f, lac_errlreg ());
    }
  fprintf (f, "\")\n");
}

void
sigint (int sig)
{
  raise_exception ("Interrupted", NIL);
}

#define TDIFF_SEC(_a, _b)				\
    ((_b)->tv_usec >= (_a)->tv_usec			\
     ? (_b)->tv_sec - (_a)->tv_sec			\
     : (_b)->tv_sec - (_a)->tv_sec - 1)

#define TDIFF_USEC(_a, _b)						\
    ((_b)->tv_usec >= t1.tv_usec					\
     ? (_b)->tv_usec - (_a)->tv_usec					\
     : (_b)->tv_usec + 1000000L - (_a)->tv_usec)


int
repl (FILE * infd, FILE * outfd, FILE * errfd)
{
  int r;
  char *buf = NULL;
  lreg_t res = NIL;
  struct timeval t1, t2;

restart:
  lac_on_error ({
      lac_error_print (errfd); res = NIL; goto restart;
    });


  do
    {
      buf = readline ("LAC>");
      if (buf == NULL)
	return -1;
      if (*buf != '\0')
	add_history (buf);

      res = sexpr_parse_string (buf,
				(lreg_t (*)(lreg_t, void *))eval,
				(void *)null_env);
      if (isatty (fileno (outfd)))
	{
	  fprintf (outfd, "=> ");
	  sexpr_fprint (outfd, res);
	  fprintf (outfd, "\n");
	}
    }
  while (1);

  lac_off_error ();
  return r;
}

int
main ()
{

  signal (SIGINT, sigint);

  lac_on_error ({
      lac_error_print (stderr); exit (-1);
    });

  null_env = lac_init ();
  repl (stdin, stdout, stderr);
  fprintf (stdout, "\ngoodbye!\n");
  return 0;
}
